/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Routines to maintain a session database in a file */

#include "backtalk.h"
#include "sysdict.h"

#ifdef ID_SESSION	/* This is all used only with cookie authentication */
char *showopt_session_module= "sess_file (file="SESSION_FILE")";

#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
extern char *crypt (char *key, char *salt);

#include "sess.h"
#include "lock.h"
#include "udb.h"


struct sess_head {
    char salt[2];		/* Last salt used */
    unsigned long seed;		/* Last seed used */
    };

struct sess_entry {
    char id[SESSION_ID_LEN];	/* Session id string */
    char user[MAX_LOGIN_LEN];	/* User name */
    time_t time;		/* Last access time */
    unsigned long ip;		/* User's IP number */
    };


int session_fd= -1;


/* OPEN_SESSION_FILE - Open the session database.  Bombs on failure.
 */

void open_session_file()
{
    if (session_fd >= 0)
	return;

    if ((session_fd= open(SESSION_FILE,O_RDWR|O_CREAT,0600)) < 0)
	die("Cannot open "SESSION_FILE);
}


/* CLOSE_SESSION_FILE - Close the session database.
 */

void close_session_file()
{
    if (session_fd >= 0)
	close(session_fd);
    session_fd= -1;
}


/* INC_SALT - Add one to a salt character.  Return true if we overflowed,
 * zero otherwise.
 */

int inc_salt(char *salt)
{
    switch (*salt)
    {
    case 'Z':  *salt= 'a'; return 0;
    case 'z':  *salt= '0'; return 0;
    case '9':  *salt= '.'; return 0;
    case '.':  *salt= '/'; return 0;
    case '/':  *salt= 'A'; return 1;
    default:   (*salt)++;  return 0;
    }
}


/* MAKE_SESSION_ID - This generates a 13-character session ID for a new user.
 * The objective here is to construct a string that is (1) likely to be
 * unique among current users, and (2) makes it hard to guess anyone else's
 * session number without actually snooping it.
 *
 * So what we do is construct a "password" based on various data that should
 * be hard to guess for other users and encrypt that with crypt().  We cycle
 * "salt" values sequentially, so the first one issued has salt "AA" and the
 * next "AB" and so on.  This forms the first two characters of the session
 * id and won't repeat for 4,096 users.  This makes having two simultaneous
 * users with the same salt (much less the same session id) pretty unlikely.
 *
 * The "password" is constructed by taking 28 bits from the sum of all
 * login-times of all users since the last time that the session file was
 * emptied.  If this file is kept unreadable, the value should be pretty hard
 * to guess.  14 bits are taken from the current second (or millesecond, if
 * available).  The other 14 bits are taken from the process id number, which
 * isn't terribly hard to guess for anyone with an account on the server, but
 * helps blur things a bit more anyway.
 *
 * Since only the encrypted version of the "passwd" is given back to users
 * as a session id, they should not be able to figure out any of the numbers
 * that went into it (which would put them a leg up on guessing what session id
 * the next user might be assigned).  Programs like "crack" won't be very
 * useful in cracking these, since the passwords aren't going to land in any
 * dictionary.  You could crack them with a supercomputer, and then have some
 * data that might help you toward guessing the next session id's but session
 * id's are only good for a few hours anyway, so you'd probably be too late.
 *
 * I'm sure better things could be done, and I'm open to suggestions, but I
 * think this should supply quite adequate security for our needs.
 */

char *make_session_id()
{
    struct sess_head shead;
    pid_t pid= getpid();
    char raw[8];
    static char *id;
    unsigned long now;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    gettimeofday(&tv,NULL);
    now= tv.tv_sec + tv.tv_usec;
#else
    now= time(0L);
#endif

    open_session_file();
    lseek(session_fd, 0L, SEEK_SET);	/* rewind */

    lock_exclusive(session_fd,SESSION_FILE);
    if (read(session_fd, &shead, sizeof(struct sess_head)) !=
	    sizeof(struct sess_head))
    {
	/* Restart the header */
	shead.salt[0]= shead.salt[1]= 'A';
	shead.seed= now;
    }
    else
    {
	/* Increment the salt */
	if (inc_salt(shead.salt)) inc_salt(shead.salt+1);
	/* Add stuff to the seed */
	shead.seed+= now;
    }

    lseek(session_fd, 0L, SEEK_SET);	/* rewind */

    write(session_fd, &shead, sizeof(struct sess_head));
    unlock_active(session_fd,SESSION_FILE);

    /* Collect 56 bits of gunk to use as our "passwd" */
    raw[0]= shead.seed % 128;
    raw[1]= (shead.seed >> 7) % 128;
    raw[2]= (shead.seed >> 14) % 128;
    raw[3]= (shead.seed >> 21) % 128;
    raw[4]= now % 128;
    raw[5]= (now >> 7) % 128;
    raw[6]= pid % 128;
    raw[7]= (pid >> 7) % 128;

    /* encrypt it to make a session id */
    id= crypt(raw,shead.salt);

    return id;
}


/* NEW_SESSION - Create new session for the given user.  Pass in a encrypted
 * password and the ip address (in ascii numbers-and-dots form) he connected
 * from.  You should already have checked that it is a valid login.  This will
 * create the session entry in the session file and return the session id
 * in malloced memory.  The caller should set the cookie.
 */

char *new_session(char *user, char *ip)
{
    char *sid;
    struct sess_entry sess, new_sess;
    time_t now= time(0L);
    long seek,offset;
    unsigned long bip;

    /* Convert IP address to binary */
    if ((new_sess.ip= inet_addr(ip)) == -1)
	die("new_session: bad ip address %s",ip);

    /* Generate the new session id and make some a copy */
    sid= strdup(make_session_id());
    /* Note: We rely on make_session_id to open the session file */

    /* Make the new session table entry */
    strncpy(new_sess.user, user, MAX_LOGIN_LEN);
    new_sess.time= now;

    offset= lseek(session_fd, sizeof(struct sess_head), SEEK_SET);

    lock_exclusive(session_fd,SESSION_FILE);

    /* Scan through file looking for an existing slot we can reuse.  We
     * always read the whole file, because we are also checking for
     * any duplicates of the session id we found.
     */

    seek= 0L;
    while (read(session_fd, &sess, sizeof(struct sess_entry)) ==
	    sizeof(struct sess_entry))
    {
	/* If this is the first empty slot remember this position */
	if (seek == 0L && (sess.id[0] == '\0' ||
	    now - sess.time > idict(VAR_EXPIRE_SESSION)))
		seek= offset;

	/* In the rare case that we notice a duplicate of this
	 * session id already in the file, modify ours and restart
	 * the scan.
	 */
	if (!strncmp(sid,sess.id,SESSION_ID_LEN))
	{
	    if (inc_salt(sid)) inc_salt(sid+1);
	    seek= 0L;
	    offset= lseek(session_fd, sizeof(struct sess_head), SEEK_SET);
	}
	else
	    offset+= sizeof(struct sess_entry);
    }

    /* If we found no empty slots, create a new one at the end */
    if (seek == 0L) seek= offset;

    /* Write out new session entry */
    lseek(session_fd, seek, SEEK_SET);
    strncpy(new_sess.id,sid,SESSION_ID_LEN);
    write(session_fd, &new_sess, sizeof(struct sess_entry));

    close_session_file();
    unlock(SESSION_FILE);
    return sid;
}


/* SESSION_LOOKUP - Look up the given session id in the session file.  If it
 * is there, and the process id (given here in ascii numbers-and-dots form)
 * matches, then update the last access time in the file and save the user
 * name in the "user" buffer (which should be at least MAX_LOGIN_LEN+1 bytes)
 * and return 0.  Otherwise return 1.
 *
 * Note that we do not lock the file while reading, not even a shared lock
 * (which may be implemented as an exclusive lock on some systems).  We don't
 * want this file to be a bottleneck on busy systems.  Thus there is some
 * risk that our slot will be swiped by the new_session() routine of some
 * other backtalk session between the time that we read it, and the time that
 * we update the last access time.  But since we carefully write out only
 * the new time, not the rest of the structure, we will only be extending
 * some other user's expiration time instead of our own.  This means that since
 * our own session slot is gone, we will get a timeout on the next access,
 * but we probably deserve it anyway, since our session must have expired on
 * on *this* access or else it wouldn't have been overwritten.
 */

int session_lookup(char *sid, char *ip, char *user)
{
    struct sess_entry sess;
    time_t now= time(0L);

#ifdef CHECK_SESSION_IP
    unsigned long bip;

    /* Convert ip address to binary
    if ((bip= inet_addr(ip)) == -1) break;
#endif

    /* Open file and skip header */
    open_session_file();
    lseek(session_fd, sizeof(struct sess_head), SEEK_SET);

    while (read(session_fd, &sess, sizeof(struct sess_entry)) ==
	sizeof(struct sess_entry))
    {
	if (strncmp(sid,sess.id,SESSION_ID_LEN)) continue;

#ifdef CHECK_SESSION_IP
	if (bip != sess.ip) break;
#endif

	if (sess.time + idict(VAR_EXPIRE_SESSION) < now) break;

	/* Update the time field (only) */
	sess.time= now;
	lseek(session_fd, - sizeof(struct sess_entry)+
	    ((char *)(&sess.time)-(char *)(&sess.id)),
	    SEEK_CUR);
	write(session_fd, &sess.time, sizeof(sess.time));

	strncpy(user,sess.user,MAX_LOGIN_LEN);
	user[MAX_LOGIN_LEN]= '\0';
	close_session_file();
	return 0;
    }
    close_session_file();
    return 1;
}


/* SESSION_DELETE - Delete the given session ID from the session file, if
 * it exists.  It is not an error if it does not exist.
 */

void session_delete(char *sid)
{
    struct sess_entry sess;

    /* Open file and skip header */
    open_session_file();
    lseek(session_fd, sizeof(struct sess_head), SEEK_SET);
    lock_exclusive(session_fd,SESSION_FILE);

    while (read(session_fd, &sess, sizeof(struct sess_entry)) ==
	sizeof(struct sess_entry))
    {
	if (!strncmp(sid,sess.id,SESSION_ID_LEN))
	{
	    /* Mark the session erased */
	    sess.id[0]= '\0';
	    lseek(session_fd, - sizeof(struct sess_entry), SEEK_CUR);
	    write(session_fd, &sess, sizeof(struct sess_entry));
	    break;
	}
    }
    close_session_file();
    unlock(SESSION_FILE);
}

#endif /* ID_SESSION */
